#!/usr/bin/env python
# coding=utf8

__version__ = '0.5'

from blinker.base import Namespace, Signal

namespace = Namespace()

user_id_changed = namespace.signal('user_id_changed')
user_id_reset = namespace.signal('user_id_reset')



class FlatRBAC(object):
    """Basic interface for the simplest possible role-based access control
    implementation."""

    # user:role
    def assign(self, user, role):
        raise NotImplementedError()

    def unassign(self, user, role):
        raise NotImplementedError()

    # role:permission
    def permit(self, role, permission):
        raise NotImplementedError()

    def revoke(self, role, permission):
        raise NotImplementedError()

    # checking
    def allows(self, role, permission):
        raise NotImplementedError()

    def allowed(self, user, permission):
        for role in self.get_assigned_roles(user):
            if self.allows(role, permission):
                return True

        return False

    # reflection
    def get_assigned_roles(self, user):
        raise NotImplementedError()


class SessionMixin(object):
    def create_session(self, user):
        raise NotImplementedError()

    # modification of session
    def activate(self, session, permissions=None):
        raise NotImplementedError()

    def deactivate(self, session, permissions=None):
        raise NotImplementedError()

    # checking
    def authorized(self, session, permissions):
        raise NotImplementedError()


class DictRBAC(FlatRBAC):
    def __init__(self):
        self._role_map = {}
        self._permission_map = {}

    def assign(self, user, role):
        self._role_map.setdefault(user, set()).add(role)

    def unassign(self, user, role):
        self._role_map.get(user, set()).discard(role)

    def permit(self, role, permission):
        self._permission_map.setdefault(role, set()).add(permission)

    def revoke(self, role, permission):
        self._permission_map.get(role, set()).discard(permission)

    def allows(self, role, permission):
        return permission in self._permission_map.get(role, set())

    def get_assigned_roles(self, user):
        return self._role_map.get(user, set())



from __future__ import absolute_import

from sqlalchemy import Column, ForeignKey, Table
from sqlalchemy.orm import relationship

from . import FlatRBAC


def _pkey_cols(decl_type):
    return decl_type.__table__.primary_key.columns.values()


def _pkey_1col(decl_type):
    cols = _pkey_cols(decl_type)

    if len(cols) != 1:
        raise TypeError('SQLAlchemyRBAC only supports models with a single '
                        'primary key column.')

    return cols[0]


class SQLAlchemyRBAC(FlatRBAC):
    """An declarative SQLAlchemy-based RBAC implementation.

    The SQLAlchemyRBAC is part of the schema and is passed three other
    declarative SQLAlchemy classes, one acting as User objects, one acting as
    Role objects and one acting as Permission objects.

    Upon instantiation, :class:`~alcohol.rbac.sqlalchemy.SQLAlchemyRBAC` will
    add the required tables for mapping users to roles and roles to permissions
    to the same :class:`~sqlalchemy.schema.MetaData` object, so
    :meth:`~sqlalchemy.schema.MetaData.create_all` will work normally, provided
    the :class:`~alcohol.rbac.sqlalchemy.SQLAlchemyRBAC` instance is created
    first.

    :note: Currently, SQLAlchemyRBAC only supports declarative models for
           user, role and permissions that have a single column primary key.

    :param user_type: A declarative SQLAlchemy model that will act as users.
    :param role_type: A declarative SQLAlchemy model that will act as roles.
    :param permission_type: A declarative SQLAlchemy model what will act as
                            permissions.
    :param prefix: A prefix for all tables generated by this RBAC.
    :param roles_lazy: The ``lazy`` argument for the
                       :func:`~sqlalchemy.orm.relationship` between users and
                       roles.
    :param permissions_lazy: The ``lazy`` argument for the
                             :func:`~sqlalchemy.orm.relationship` between
                             roles and permissions.
    """

    def __init__(self,
                 user_type,
                 role_type,
                 permission_type,
                 prefix='rbac_',
                 roles_lazy='joined',
                 permissions_lazy='joined', ):
        if not (role_type.metadata == permission_type.metadata ==
                user_type.metadata):
            raise TypeError('All three models must be part of the same '
                            'metadata.')

        metadata = user_type.metadata

        self.prefix = prefix
        self._roles_rel = '_' + self.prefix + 'roles'
        self._perms_rel = '_' + self.prefix + 'permissions'

        user_key_col = _pkey_1col(user_type)
        role_key_col = _pkey_1col(role_type)
        permission_key_col = _pkey_1col(permission_type)

        user_role_map = Table(self.prefix + 'user_role_map',
                              metadata,
                              Column('user_pkey', user_key_col.type,
                                     ForeignKey(user_key_col)),
                              Column('role_pkey', role_key_col.type,
                                     ForeignKey(role_key_col)), )

        role_permissions_map = Table(self.prefix + 'role_permission_map',
                                     metadata,
                                     Column('role_pkey', role_key_col.type,
                                            ForeignKey(role_key_col)),
                                     Column('permission_pkey',
                                            permission_key_col.type,
                                            ForeignKey(permission_key_col)), )

        # add orm relationships
        setattr(user_type,
                self._roles_rel,
                relationship(role_type,
                             secondary=user_role_map,
                             lazy=roles_lazy, ))
        setattr(role_type,
                self._perms_rel,
                relationship(permission_type,
                             secondary=role_permissions_map,
                             lazy=permissions_lazy, ))

    # RBAC api:
    def assign(self, user, role):
        getattr(user, self._roles_rel).append(role)

    def unassign(self, user, role):
        try:
            getattr(user, self._roles_rel).remove(role)
        except ValueError:
            pass  # not in list, ignore

    def permit(self, role, permission):
        getattr(role, self._perms_rel).append(permission)

    def revoke(self, role, permission):
        try:
            getattr(role, self._perms_rel).remove(permission)
        except ValueError:
            pass  # not in list, ignore

    def allows(self, role, permission):
        return permission in getattr(role, self._perms_rel)

    def get_assigned_roles(self, user):
        return list(getattr(user, self._roles_rel))



#!/usr/bin/env python
# coding=utf8

from __future__ import absolute_import
from datetime import datetime
from sqlalchemy import Column, String, Unicode, DateTime
from sqlalchemy.sql.expression import case
from sqlalchemy.ext.hybrid import hybrid_property
from . import PasswordMixin, EmailMixin


class SQLAlchemyPasswordMixin(PasswordMixin):
    """Adds a :class:`~sqlalchemy.types.String`
    :class:`~sqlalchemy.schema.Column` containing the password hash.

    Supports the same interface as :class:`~alcohol.mixins.PasswordMixin`."""
    HASH_FIELD_LEN = 511
    _pwhash = Column(String(HASH_FIELD_LEN))


class SQLAlchemyEmailMixin(EmailMixin):
    """Adds a :class:`~sqlalchemy.types.Unicode`
    :class:`~sqlalchemy.schema.Column` named ``email`` for storing a users
    email address.

    Supports the same interface as :class:`~alcohol.mixins.EmailMixin`."""
    MAX_EMAIL_LENGTH = 1023
    email = Column(Unicode(MAX_EMAIL_LENGTH))


class TimestampMixin(object):
    """A mixin that adds two timestamp fields, `created` and `modified`. The
    `created` timestamp is updated only on creation, while every SQL UPDATE
    will trigger a refresh of the `modified` timestamp."""

    created = Column(DateTime, default=datetime.utcnow, nullable=False)
    """A :py:class:`datetime.datetime` instance containing the time this record
    was created."""

    modified = Column(DateTime, onupdate=datetime.utcnow)
    """A :py:class:`datetime.datetime` instance containing the time this record
    was last modified."""

    @hybrid_property
    def last_modified(self):
        """A hybrid property that returns the newer of ``created`` or
        ``modified``, or ``None`` if both are ``NULL``. Can be used in SQL
        expressions."""
        if self.created is not None and self.modified is not None:
            return (self.modified if self.modified >= self.created else
                    self.created)

        if self.modified:
            return self.modified

        if self.created:
            return self.created

        # both None, returns None

    @last_modified.expression
    def last_modified(cls):
        return case(
            [
                (cls.created <= cls.modified, cls.modified),
                (cls.created > cls.modified, cls.created),

                # at least one must be NULL
                (cls.created != None, cls.created),
                (cls.modified != None, cls.modified),

                # fallthrough to NULL
            ],
            else_=None)



#!/usr/bin/env python
# coding=utf8

import os

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='alcohol',
      version='0.5',
      description='Handles user authentication, in a way.',
      long_description=read('README.rst'),
      author='Marc Brinkmann',
      author_email='git@marcbrinkmann.de',
      url='http://github.com/mbr/alcohol',
      license='MIT',
      packages=find_packages(exclude=['tests']),
      # FIXME: sqlalchemy is not a strict requirement, but building the docs
      #        is not possible without right now. unleash needs to support
      #        "for docs"-dependencies
      install_requires=['blinker', 'passlib>=1.6', 'six', 'itsdangerous>=0.14',
                        'sqlalchemy'], )



#!/usr/bin/env python
# coding=utf8

from binascii import hexlify
import os

from itsdangerous import TimestampSigner, BadData, URLSafeTimedSerializer
import passlib.apps


DAY = 60 * 60 * 24


class PasswordMixin(object):
    """A mixin that stores a key based on a password. An attribute named
    `_pwhash` will be used to store the password hash."""
    crypt_context = passlib.apps.custom_app_context  # overridable default

    def check_password(self, password):
        """Check if a supplied password is the same as the user's password.

        :param password: Password to be checked.
        :return: ``True`` if valid, ``False`` otherwise.
        """
        return self.crypt_context.verify(password, self._pwhash)

    def _create_signer(self, secret_key):
        return TimestampSigner(secret_key, self._pwhash, key_derivation='hmac')

    def check_password_reset_token(self, secret_key, token, max_age_sec=DAY):
        """Checks if a supplied password-reset token is valid.

        :param secret_key: Your applications secret key.
        :param password: Password-reset token to be checked.
        :param max_age_sec: The maximum age in seconds this token may be old
                            before its considered expired. Default is 24 hours.
        :return: ``True`` if valid, ``False`` otherwise.
        """

        signer = self._create_signer(secret_key)
        try:
            signer.unsign(token, max_age=max_age_sec)
            return True
        except BadData:
            return False

    def create_reset_password_token(self, secret_key, random_source=os.urandom,
                                    nonce_size=5):
        """Create a signed password reset token.

        A pasword reset token using a key derived from ``secret_key`` and
        the current password hash, causing it to stop working once the password
        has been altered.

        It also includes a nonce, so that attackers cannot tell whether a
        password-reset request has been made twice for the same password.

        :param secret_key: The application's own secret key.
        :param random_source: The random source to use to create the nonce.
                              Defaults to :func:`os.urandom`.
        :param nonce_size: Number of bytes in the nonce. Each additional byte
                           will increase the resulting tokens length by 2.
        :return: An urlsafe string.
        """

        # sign a few random bytes to hide repetitions
        signer = self._create_signer(secret_key)
        return signer.sign(hexlify(random_source(nonce_size)))

    @property
    def password(self):
        """The users password. This is a write-only property, attempting to
        read it will throw an exception. Use this to set the users password."""
        raise TypeError(
            'password property is write-only, use check_password'
        )

    @password.setter
    def password(self, new_password):
        self._pwhash = self.crypt_context.encrypt(new_password)


class EmailMixin(object):
    """Adds an ``email`` attribute and supports generating email activation
    tokens."""

    email = None
    """An email address. Not validated in any form."""

    def _create_serializer(self, secret_key):
        return URLSafeTimedSerializer(
            secret_key, self.email, signer_kwargs={'key_derivation': 'hmac'}
        )

    def activate_email(self, secret_key, token, max_age_sec=DAY):
        """Checks if the email activation token is valid. If it is, updates the
        users email address with the one saved in the token.

        :param secret_key: The application's own secret key.
        :param token: The activation token.
        :param max_age_sec: The maximum age in seconds this token may be old
                            before its considered expired. Default is 24 hours.

        :return: ``True`` if the activation was successful, ``False``
                otherwise."""

        serializer = self._create_serializer(secret_key)
        try:
            self.email = serializer.loads(token, max_age=max_age_sec)
            return True
        except BadData:
            return False

    def create_email_activation_token(self, secret_key, email):
        """Creates a new activation token that allows changing the email
        address. The token will tied to the old email address and works only
        if the address has not changed in the meantime.

        :param secret_key: The application's own secret key.
        :param email: The desired new email address. Will be encoded inside
                      the token.
        :return: An urlsafe string.
        """

        return self._create_serializer(secret_key).dumps(email)
